@@ -24,18 +24,19 @@
 // IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 // DEALINGS IN THE SOFTWARE.
 
-use crate::server::helpers::{send_error, send_response};
+use crate::server::helpers::{send_call_error, send_error, send_response};
 use crate::server::resource_limiting::{ResourceGuard, ResourceTable, ResourceVec, Resources};
 use beef::Cow;
 use futures_channel::{mpsc, oneshot};
 use futures_util::{future::BoxFuture, FutureExt, StreamExt};
+use jsonrpsee_types::to_json_raw_value;
+use jsonrpsee_types::v2::error::{invalid_subscription_err, CALL_EXECUTION_FAILED_CODE};
 use jsonrpsee_types::{
-	error::{CallError, Error, SubscriptionClosedError},
+	error::{Error, SubscriptionClosedError},
 	traits::ToRpcParams,
 	v2::{
-		error::{CALL_EXECUTION_FAILED_CODE, UNKNOWN_ERROR_CODE},
-		ErrorCode, ErrorObject, Id, Params, Request, Response, SubscriptionId as RpcSubscriptionId,
-		SubscriptionPayload, SubscriptionResponse, TwoPointZero,
+		ErrorCode, Id, Params, Request, Response, SubscriptionId as RpcSubscriptionId, SubscriptionPayload,
+		SubscriptionResponse, TwoPointZero,
 	},
 	DeserializeOwned,
 };
@@ -54,10 +55,11 @@ use std::sync::Arc;
 /// implemented as a function pointer to a `Fn` function taking four arguments:
 /// the `id`, `params`, a channel the function uses to communicate the result (or error)
 /// back to `jsonrpsee`, and the connection ID (useful for the websocket transport).
-pub type SyncMethod = Arc<dyn Send + Sync + Fn(Id, Params, &MethodSink, ConnectionId)>;
+pub type SyncMethod = Arc<dyn Send + Sync + Fn(Id, Params, &MethodSink, ConnectionId, MaxResponseSize)>;
 /// Similar to [`SyncMethod`], but represents an asynchronous handler and takes an additional argument containing a [`ResourceGuard`] if configured.
-pub type AsyncMethod<'a> =
-	Arc<dyn Send + Sync + Fn(Id<'a>, Params<'a>, MethodSink, ConnectionId, Option<ResourceGuard>) -> BoxFuture<'a, ()>>;
+pub type AsyncMethod<'a> = Arc<
+	dyn Send + Sync + Fn(Id<'a>, Params<'a>, MethodSink, Option<ResourceGuard>, MaxResponseSize) -> BoxFuture<'a, ()>,
+>;
 /// Connection ID, used for stateful protocol such as WebSockets.
 /// For stateless protocols such as http it's unused, so feel free to set it some hardcoded value.
 pub type ConnectionId = usize;
@@ -65,6 +67,8 @@ pub type ConnectionId = usize;
 pub type SubscriptionId = u64;
 /// Sink that is used to send back the result to the server for a specific method.
 pub type MethodSink = mpsc::UnboundedSender<String>;
+/// Max response size in bytes for a executed call.
+pub type MaxResponseSize = u32;
 
 type Subscribers = Arc<Mutex<FxHashMap<SubscriptionKey, (MethodSink, oneshot::Receiver<()>)>>>;
 
@@ -147,19 +151,20 @@ impl MethodCallback {
 		req: Request<'_>,
 		conn_id: ConnectionId,
 		claimed: Option<ResourceGuard>,
+		max_response_size: MaxResponseSize,
 	) -> Option<BoxFuture<'static, ()>> {
 		let id = req.id.clone();
 		let params = Params::new(req.params.map(|params| params.get()));
 
 		match &self.callback {
 			MethodKind::Sync(callback) => {
-				log::trace!(
+				tracing::trace!(
 					"[MethodCallback::execute] Executing sync callback, params={:?}, req.id={:?}, conn_id={:?}",
 					params,
 					id,
 					conn_id
 				);
-				(callback)(id, params, tx, conn_id);
+				(callback)(id, params, tx, conn_id, max_response_size);
 
 				// Release claimed resources
 				drop(claimed);
@@ -170,14 +175,14 @@ impl MethodCallback {
 				let tx = tx.clone();
 				let params = params.into_owned();
 				let id = id.into_owned();
-				log::trace!(
+				tracing::trace!(
 					"[MethodCallback::execute] Executing async callback, params={:?}, req.id={:?}, conn_id={:?}",
 					params,
 					id,
 					conn_id
 				);
 
-				Some((callback)(id, params, tx, conn_id, claimed))
+				Some((callback)(id, params, tx, claimed, max_response_size))
 			}
 		}
 	}
@@ -284,10 +289,16 @@ impl Methods {
 	}
 
 	/// Attempt to execute a callback, sending the resulting JSON (success or error) to the specified sink.
-	pub fn execute(&self, tx: &MethodSink, req: Request, conn_id: ConnectionId) -> Option<BoxFuture<'static, ()>> {
-		log::trace!("[Methods::execute] Executing request: {:?}", req);
+	pub fn execute(
+		&self,
+		tx: &MethodSink,
+		req: Request,
+		conn_id: ConnectionId,
+		max_response_size: MaxResponseSize,
+	) -> Option<BoxFuture<'static, ()>> {
+		tracing::trace!("[Methods::execute] Executing request: {:?}", req);
 		match self.callbacks.get(&*req.method) {
-			Some(callback) => callback.execute(tx, req, conn_id, None),
+			Some(callback) => callback.execute(tx, req, conn_id, None, max_response_size),
 			None => {
 				send_error(req.id, tx, ErrorCode::MethodNotFound.into());
 				None
@@ -302,13 +313,14 @@ impl Methods {
 		req: Request,
 		conn_id: ConnectionId,
 		resources: &Resources,
+		max_response_size: MaxResponseSize,
 	) -> Option<BoxFuture<'static, ()>> {
-		log::trace!("[Methods::execute_with_resources] Executing request: {:?}", req);
+		tracing::trace!("[Methods::execute_with_resources] Executing request: {:?}", req);
 		match self.callbacks.get(&*req.method) {
 			Some(callback) => match callback.claim(&req.method, resources) {
-				Ok(guard) => callback.execute(tx, req, conn_id, Some(guard)),
+				Ok(guard) => callback.execute(tx, req, conn_id, Some(guard), max_response_size),
 				Err(err) => {
-					log::error!("[Methods::execute_with_resources] failed to lock resources: {:?}", err);
+					tracing::error!("[Methods::execute_with_resources] failed to lock resources: {:?}", err);
 					send_error(req.id, tx, ErrorCode::ServerIsBusy.into());
 					None
 				}
@@ -340,7 +352,7 @@ impl Methods {
 
 		let (tx, mut rx) = mpsc::unbounded();
 
-		if let Some(fut) = self.execute(&tx, req, 0) {
+		if let Some(fut) = self.execute(&tx, req, 0, MaxResponseSize::MAX) {
 			fut.await;
 		}
 
@@ -351,13 +363,13 @@ impl Methods {
 	/// [`SubscriptionId`] and a channel on which subscription JSON payloads can be received.
 	pub async fn test_subscription(&self, method: &str, params: impl ToRpcParams) -> TestSubscription {
 		let params = params.to_rpc_params().expect("valid JSON-RPC params");
-		log::trace!("[Methods::test_subscription] Calling subscription method: {:?}, params: {:?}", method, params);
+		tracing::trace!("[Methods::test_subscription] Calling subscription method: {:?}, params: {:?}", method, params);
 		let req =
 			Request { jsonrpc: TwoPointZero, id: Id::Number(0), method: Cow::borrowed(method), params: Some(&params) };
 
 		let (tx, mut rx) = mpsc::unbounded();
 
-		if let Some(fut) = self.execute(&tx, req, 0) {
+		if let Some(fut) = self.execute(&tx, req, 0, MaxResponseSize::MAX) {
 			fut.await;
 		}
 		let response = rx.next().await.expect("Could not establish subscription.");
@@ -424,35 +436,10 @@ impl<Context: Send + Sync + 'static> RpcModule<Context> {
 		let ctx = self.ctx.clone();
 		let callback = self.methods.verify_and_insert(
 			method_name,
-			MethodCallback::new_sync(Arc::new(move |id, params, tx, _| {
+			MethodCallback::new_sync(Arc::new(move |id, params, tx, _, max_response_size| {
 				match callback(params, &*ctx) {
-					Ok(res) => send_response(id, tx, res),
-					Err(Error::Call(CallError::InvalidParams(e))) => {
-						let error = ErrorObject { code: ErrorCode::InvalidParams, message: &e.to_string(), data: None };
-						send_error(id, tx, error)
-					}
-					Err(Error::Call(CallError::Failed(e))) => {
-						let err = ErrorObject {
-							code: ErrorCode::ServerError(CALL_EXECUTION_FAILED_CODE),
-							message: &e.to_string(),
-							data: None,
-						};
-						send_error(id, tx, err)
-					}
-					Err(Error::Call(CallError::Custom { code, message, data })) => {
-						let err = ErrorObject { code: code.into(), message: &message, data: data.as_deref() };
-						send_error(id, tx, err)
-					}
-					// This should normally not happen because the most common use case is to
-					// return `Error::Call` in `register_method`.
-					Err(e) => {
-						let err = ErrorObject {
-							code: ErrorCode::ServerError(UNKNOWN_ERROR_CODE),
-							message: &e.to_string(),
-							data: None,
-						};
-						send_error(id, tx, err)
-					}
+					Ok(res) => send_response(id, tx, res, max_response_size),
+					Err(err) => send_call_error(id, tx, err),
 				};
 			})),
 		)?;
@@ -474,38 +461,12 @@ impl<Context: Send + Sync + 'static> RpcModule<Context> {
 		let ctx = self.ctx.clone();
 		let callback = self.methods.verify_and_insert(
 			method_name,
-			MethodCallback::new_async(Arc::new(move |id, params, tx, _conn_id, claimed| {
+			MethodCallback::new_async(Arc::new(move |id, params, tx, claimed, max_response_size| {
 				let ctx = ctx.clone();
 				let future = async move {
 					match callback(params, ctx).await {
-						Ok(res) => send_response(id, &tx, res),
-						Err(Error::Call(CallError::InvalidParams(e))) => {
-							let error =
-								ErrorObject { code: ErrorCode::InvalidParams, message: &e.to_string(), data: None };
-							send_error(id, &tx, error)
-						}
-						Err(Error::Call(CallError::Failed(e))) => {
-							let err = ErrorObject {
-								code: ErrorCode::ServerError(CALL_EXECUTION_FAILED_CODE),
-								message: &e.to_string(),
-								data: None,
-							};
-							send_error(id, &tx, err)
-						}
-						Err(Error::Call(CallError::Custom { code, message, data })) => {
-							let err = ErrorObject { code: code.into(), message: &message, data: data.as_deref() };
-							send_error(id, &tx, err)
-						}
-						// This should normally not happen because the most common use case is to
-						// return `Error::Call` in `register_async_method`.
-						Err(e) => {
-							let err = ErrorObject {
-								code: ErrorCode::ServerError(UNKNOWN_ERROR_CODE),
-								message: &e.to_string(),
-								data: None,
-							};
-							send_error(id, &tx, err)
-						}
+						Ok(res) => send_response(id, &tx, res, max_response_size),
+						Err(err) => send_call_error(id, &tx, err),
 					};
 
 					// Release claimed resources
@@ -518,9 +479,57 @@ impl<Context: Send + Sync + 'static> RpcModule<Context> {
 		Ok(MethodResourcesBuilder { build: ResourceVec::new(), callback })
 	}
 
-	/// Register a new RPC subscription that invokes callback on every subscription request.
-	/// The callback itself takes three parameters:
-	///     - [`Params`]: JSONRPC parameters in the subscription request.
+	/// Register a new **blocking** synchronous RPC method, which computes the response with the given callback.
+	/// Unlike the regular [`register_method`](RpcModule::register_method), this method can block its thread and perform expensive computations.
+	pub fn register_blocking_method<R, F>(
+		&mut self,
+		method_name: &'static str,
+		callback: F,
+	) -> Result<MethodResourcesBuilder, Error>
+	where
+		Context: Send + Sync + 'static,
+		R: Serialize,
+		F: Fn(Params, Arc<Context>) -> Result<R, Error> + Copy + Send + Sync + 'static,
+	{
+		let ctx = self.ctx.clone();
+		let callback = self.methods.verify_and_insert(
+			method_name,
+			MethodCallback::new_async(Arc::new(move |id, params, tx, claimed, max_response_size| {
+				let ctx = ctx.clone();
+
+				tokio::task::spawn_blocking(move || {
+					match callback(params, ctx) {
+						Ok(res) => send_response(id, &tx, res, max_response_size),
+						Err(err) => send_call_error(id, &tx, err),
+					};
+
+					// Release claimed resources
+					drop(claimed);
+				})
+				.map(|err| {
+					tracing::error!("Join error for blocking RPC method: {:?}", err);
+				})
+				.boxed()
+			})),
+		)?;
+
+		Ok(MethodResourcesBuilder { build: ResourceVec::new(), callback })
+	}
+
+	/// Register a new RPC subscription that invokes s callback on every subscription call.
+	///
+	/// This method ensures that the `subscription_method_name` and `unsubscription_method_name` are unique.
+	/// The `notif_method_name` argument sets the content of the `method` field in the JSON document that
+	/// the server sends back to the client. The uniqueness of this value is not machine checked and it's up to
+	/// the user to ensure it is not used in any other [`RpcModule`] used in the server.
+	///
+	/// # Arguments
+	///
+	/// * `subscription_method_name` - name of the method to call to initiate a subscription
+	/// * `notif_method_name` - name of method to be used in the subscription payload (technically a JSON-RPC notification)
+	/// * `unsubscription_method` - name of the method to call to terminate a subscription
+	/// *  `callback` - A callback to invoke on each subscription; it takes three parameters:
+	///     - [`Params`]: JSON-RPC parameters in the subscription call.
 	///     - [`SubscriptionSink`]: A sink to send messages to the subscriber.
 	///     - Context: Any type that can be embedded into the [`RpcModule`].
 	///
@@ -531,7 +540,7 @@ impl<Context: Send + Sync + 'static> RpcModule<Context> {
 	/// use jsonrpsee_utils::server::rpc_module::RpcModule;
 	///
 	/// let mut ctx = RpcModule::new(99_usize);
-	/// ctx.register_subscription("sub", "unsub", |params, mut sink, ctx| {
+	/// ctx.register_subscription("sub", "notif_name", "unsub", |params, mut sink, ctx| {
 	///     let x: usize = params.one()?;
 	///     std::thread::spawn(move || {
 	///         let sum = x + (*ctx);
@@ -543,6 +552,7 @@ impl<Context: Send + Sync + 'static> RpcModule<Context> {
 	pub fn register_subscription<F>(
 		&mut self,
 		subscribe_method_name: &'static str,
+		notif_method_name: &'static str,
 		unsubscribe_method_name: &'static str,
 		callback: F,
 	) -> Result<(), Error>
@@ -556,15 +566,15 @@ impl<Context: Send + Sync + 'static> RpcModule<Context> {
 
 		self.methods.verify_method_name(subscribe_method_name)?;
 		self.methods.verify_method_name(unsubscribe_method_name)?;
-		let ctx = self.ctx.clone();
 
+		let ctx = self.ctx.clone();
 		let subscribers = Subscribers::default();
 
 		{
 			let subscribers = subscribers.clone();
 			self.methods.mut_callbacks().insert(
 				subscribe_method_name,
-				MethodCallback::new_sync(Arc::new(move |id, params, method_sink, conn_id| {
+				MethodCallback::new_sync(Arc::new(move |id, params, method_sink, conn_id, max_response_size| {
 					let (conn_tx, conn_rx) = oneshot::channel::<()>();
 					let sub_id = {
 						const JS_NUM_MASK: SubscriptionId = !0 >> 11;
@@ -576,23 +586,23 @@ impl<Context: Send + Sync + 'static> RpcModule<Context> {
 						sub_id
 					};
 
-					send_response(id.clone(), method_sink, sub_id);
+					send_response(id.clone(), method_sink, sub_id, max_response_size);
 
 					let sink = SubscriptionSink {
 						inner: method_sink.clone(),
-						method: subscribe_method_name,
+						method: notif_method_name,
 						subscribers: subscribers.clone(),
 						uniq_sub: SubscriptionKey { conn_id, sub_id },
 						is_connected: Some(conn_tx),
 					};
 					if let Err(err) = callback(params, sink, ctx.clone()) {
-						log::error!(
+						tracing::error!(
 							"subscribe call '{}' failed: {:?}, request id={:?}",
 							subscribe_method_name,
 							err,
 							id
 						);
-						send_error(id, method_sink, ErrorCode::ServerError(-1).into());
+						send_error(id, method_sink, ErrorCode::ServerError(CALL_EXECUTION_FAILED_CODE).into());
 					}
 				})),
 			);
@@ -601,21 +611,28 @@ impl<Context: Send + Sync + 'static> RpcModule<Context> {
 		{
 			self.methods.mut_callbacks().insert(
 				unsubscribe_method_name,
-				MethodCallback::new_sync(Arc::new(move |id, params, tx, conn_id| {
+				MethodCallback::new_sync(Arc::new(move |id, params, tx, conn_id, max_response_size| {
 					let sub_id = match params.one() {
 						Ok(sub_id) => sub_id,
 						Err(_) => {
-							log::error!(
-								"unsubscribe call '{}' failed: couldn't parse subscription id, request id={:?}",
+							tracing::error!(
+								"unsubscribe call '{}' failed: couldn't parse subscription id={:?} request id={:?}",
 								unsubscribe_method_name,
+								params,
 								id
 							);
-							send_error(id, tx, ErrorCode::ServerError(-1).into());
+							let err = to_json_raw_value(&"Invalid subscription ID type, must be integer").ok();
+							send_error(id, tx, invalid_subscription_err(err.as_deref()));
 							return;
 						}
 					};
-					subscribers.lock().remove(&SubscriptionKey { conn_id, sub_id });
-					send_response(id, tx, "Unsubscribed");
+
+					if subscribers.lock().remove(&SubscriptionKey { conn_id, sub_id }).is_some() {
+						send_response(id, tx, "Unsubscribed", max_response_size);
+					} else {
+						let err = to_json_raw_value(&format!("Invalid subscription ID={}", sub_id)).ok();
+						send_error(id, tx, invalid_subscription_err(err.as_deref()))
+					}
 				})),
 			);
 		}
@@ -703,6 +720,7 @@ impl SubscriptionSink {
 	fn inner_close(&mut self, err: &SubscriptionClosedError) {
 		self.is_connected.take();
 		if let Some((sink, _)) = self.subscribers.lock().remove(&self.uniq_sub) {
+			tracing::debug!("Closing subscription: {:?}", self.uniq_sub.sub_id);
 			let msg = self.build_message(err).expect("valid json infallible; qed");
 			let _ = sink.unbounded_send(msg);
 		}
@@ -735,13 +753,17 @@ impl TestSubscription {
 		self.sub_id
 	}
 
-	/// Get the next element of type T from the underlying stream.
+	/// Returns `Some((val, sub_id))` for the next element of type T from the underlying stream,
+	/// otherwise `None` if the subscruption was closed.
 	///
-	/// Panics if the stream was closed or if the decoding the value as `T`.
-	pub async fn next<T: DeserializeOwned>(&mut self) -> (T, jsonrpsee_types::v2::SubscriptionId) {
-		let raw = self.rx.next().await.expect("subscription not closed");
-		let val: SubscriptionResponse<T> = serde_json::from_str(&raw).expect("valid response");
-		(val.params.result, val.params.subscription)
+	/// # Panics
+	///
+	/// If the decoding the value as `T` fails.
+	pub async fn next<T: DeserializeOwned>(&mut self) -> Option<(T, jsonrpsee_types::v2::SubscriptionId)> {
+		let raw = self.rx.next().await?;
+		let val: SubscriptionResponse<T> =
+			serde_json::from_str(&raw).expect("valid response in TestSubscription::next()");
+		Some((val.params.result, val.params.subscription))
 	}
 }
 
@@ -776,7 +798,7 @@ mod tests {
 	fn rpc_context_modules_can_register_subscriptions() {
 		let cx = ();
 		let mut cxmodule = RpcModule::new(cx);
-		let _subscription = cxmodule.register_subscription("hi", "goodbye", |_, _, _| Ok(()));
+		let _subscription = cxmodule.register_subscription("hi", "hi", "goodbye", |_, _, _| Ok(()));
 
 		assert!(cxmodule.method("hi").is_some());
 		assert!(cxmodule.method("goodbye").is_some());
@@ -914,10 +936,10 @@ mod tests {
 	async fn subscribing_without_server() {
 		let mut module = RpcModule::new(());
 		module
-			.register_subscription("my_sub", "my_unsub", |_, mut sink, _| {
+			.register_subscription("my_sub", "my_sub", "my_unsub", |_, mut sink, _| {
 				let mut stream_data = vec!['0', '1', '2'];
 				std::thread::spawn(move || loop {
-					log::debug!("This is your friendly subscription sending data.");
+					tracing::debug!("This is your friendly subscription sending data.");
 					if let Some(letter) = stream_data.pop() {
 						if let Err(Error::SubscriptionClosed(_)) = sink.send(&letter) {
 							return;
@@ -933,14 +955,39 @@ mod tests {
 
 		let mut my_sub: TestSubscription = module.test_subscription("my_sub", Vec::<()>::new()).await;
 		for i in (0..=2).rev() {
-			let (val, id) = my_sub.next::<char>().await;
+			let (val, id) = my_sub.next::<char>().await.unwrap();
 			assert_eq!(val, std::char::from_digit(i, 10).unwrap());
 			assert_eq!(id, v2::params::SubscriptionId::Num(my_sub.subscription_id()));
 		}
 
-		// The subscription is now closed
-		let (sub_closed_err, _) = my_sub.next::<SubscriptionClosedError>().await;
+		// The subscription is now closed by the server.
+		let (sub_closed_err, _) = my_sub.next::<SubscriptionClosedError>().await.unwrap();
 		assert_eq!(sub_closed_err.subscription_id(), my_sub.subscription_id());
 		assert_eq!(sub_closed_err.close_reason(), "Closed by the server");
 	}
+
+	#[tokio::test]
+	async fn close_test_subscribing_without_server() {
+		let mut module = RpcModule::new(());
+		module
+			.register_subscription("my_sub", "my_sub", "my_unsub", |_, mut sink, _| {
+				std::thread::spawn(move || loop {
+					if let Err(Error::SubscriptionClosed(_)) = sink.send(&"lo") {
+						return;
+					}
+					std::thread::sleep(std::time::Duration::from_millis(500));
+				});
+				Ok(())
+			})
+			.unwrap();
+
+		let mut my_sub: TestSubscription = module.test_subscription("my_sub", Vec::<()>::new()).await;
+		let (val, id) = my_sub.next::<String>().await.unwrap();
+		assert_eq!(&val, "lo");
+		assert_eq!(id, v2::params::SubscriptionId::Num(my_sub.subscription_id()));
+
+		// close the subscription to ensure it doesn't return any items.
+		my_sub.close();
+		assert_eq!(None, my_sub.next::<String>().await);
+	}
 }
